import json
from pathlib import Path
from typing import Dict, Union

import dill as pickle
import flask
from flask import Flask, Response, request
from flask.views import MethodView

from stickybeak.handle_requests import InjectData, get_server_data, inject


def setup(app: Flask, endpoint: str, project_dir: Union[str, Path]) -> None:
    project_dir = Path(project_dir)

    class InjectView(MethodView):
        def post(self) -> Response:
            data: InjectData = pickle.loads(request.data)
            return Response(inject(data), status=200)

    class DataView(MethodView):
        def get(self) -> Response:
            data = get_server_data(project_dir)
            response_data = json.dumps(data)
            return Response(response_data, status=200)

    app.add_url_rule(f"/{endpoint}/inject", view_func=InjectView.as_view("inject"))
    app.add_url_rule(f"/{endpoint}/data", view_func=DataView.as_view("data"))



import atexit
from http.server import BaseHTTPRequestHandler, HTTPServer
from io import BytesIO
import json
from pathlib import Path
from threading import Thread
from time import sleep
from typing import Any, Optional

import dill as pickle

from stickybeak import handle_requests

__all__ = ["Server"]

from stickybeak.handle_requests import InjectData


class Server(Thread):
    project_root: Path
    port: int
    timeout: Optional[float]
    collect_source: bool

    def __init__(
        self, project_root: Optional[Path], port: int, timeout: Optional[float] = None, collect_source: bool = False
    ) -> None:
        super().__init__()

        self.project_root = project_root
        self.port = port
        self.collect_source = collect_source
        self.timeout = timeout

        class Handler(BaseHTTPRequestHandler):
            project_root = self.project_root

            def do_POST(self) -> None:
                content_length = int(self.headers["Content-Length"])
                body = self.rfile.read(content_length)

                self.send_response(200)
                self.end_headers()

                data: InjectData = pickle.loads(body)
                response = BytesIO()
                response.write(handle_requests.inject(data))
                self.wfile.write(response.getvalue())

            def do_GET(self2) -> None:
                self2.send_response(200)
                self2.send_header("content-type", "application/json")
                self2.end_headers()

                data = handle_requests.get_server_data(self2.project_root)
                response_data = json.dumps(data)
                self2.wfile.write(response_data.encode("utf-8"))

            def handle_http(self) -> None:
                return

            def respond(self) -> None:
                return

            def exit(self) -> None:
                pass

            def log_message(self, format: str, *args: Any) -> None:
                return

        self.httpd = HTTPServer(("localhost", self.port), Handler)
        self.setDaemon(True)

    def _watchdog(self) -> None:
        assert self.timeout is not None
        sleep(self.timeout)
        self.exit()

    def exit(self) -> None:
        self.httpd.shutdown()

    def run(self) -> None:
        atexit.register(self.exit)

        if self.timeout:
            Thread(target=self._watchdog).start()

        self.httpd.serve_forever()
        self.httpd.server_close()

    def is_running(self) -> bool:
        return self.is_alive()



from dataclasses import dataclass
import json
from math import ceil
from pathlib import Path
import re
import sys
from typing import Any, Dict
from urllib.parse import urljoin

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter, Response
from urllib3 import Retry


class SubSecondRetry(Retry):
    """Handle floats like: Retry-After: 0.223"""

    def parse_retry_after(self, retry_after: str) -> float:
        if re.match(r"^\s*[0-9]+\.*[0-9]+\s*$", retry_after):
            return ceil(float(retry_after))
        return super().parse_retry_after(retry_after)


@dataclass
class Client:
    base_url: str

    def __post_init__(self) -> None:
        self._retry_session = Session()

        retry = SubSecondRetry(total=6, read=6, connect=6, status=6, backoff_factor=0.1)
        self._retry_session.mount("http://", HTTPAdapter(max_retries=retry))
        self._retry_session.mount("https://", HTTPAdapter(max_retries=retry))

        self._session = Session()
        self._session.mount("http://", HTTPAdapter())
        self._session.mount("https://", HTTPAdapter())

    def get(self, url: str) -> Dict[str, Any]:
        joined_url = urljoin(self.base_url, url)
        response: Response = self._retry_session.get(joined_url)

        try:
            response.raise_for_status()
        except HTTPError:
            sys.stderr.write(response.text)
            sys.stderr.flush()
            raise

        return json.loads(response.content or "null")

    def post(self, url: str, data: bytes) -> Response:
        joined_url = urljoin(self.base_url, url)

        response: Response = self._session.post(joined_url, data=data, timeout=60)
        response.raise_for_status()
        return response


def get_site_packages_dir_from_venv(venv: Path) -> Path:
    ret = next((venv / "lib").glob("*"))
    if ret.name != "site-packages":
        ret /= "site-packages"
    return ret



import glob
import os
from pathlib import Path
import traceback
from typing import Any, Dict, List, Optional, Union

import dill as pickle
from typing_extensions import TypedDict

from stickybeak import sandbox, utils
from stickybeak.vendored.pip._internal.operations import freeze  # type: ignore
from stickybeak.vendored.pip._internal.utils.misc import get_installed_distributions

INJECT_ENDPOINT = "inject"
SERVER_DATA_ENDPOINT = "data"


class InjectData(TypedDict):
    source: str  # source code of function/class
    filename: str  # filename of original injection function/class definition
    offset: int  # injection code line offset
    call: str  # call statement
    args: List[Any]
    kwargs: Dict[str, Any]


class Requirement(TypedDict):
    project_name: str
    egg_info: str
    key: str
    version: str


class ServerData(TypedDict):
    source: Dict[str, str]  # file -> source
    requirements: Dict[str, Requirement]  # requirement name -> Requirement
    envs: Dict[str, str]  # env name -> env value


def call_function(data: InjectData) -> bytes:
    """Function where the injected code will be executed.
    Helps to avoid local variable conflicts."""

    ret: bytes
    results: Dict[str, Any] = {}

    offset = "\n" * (data["offset"] - 3)

    code = f"""
{offset}
{data["source"]}
__return = {data["call"]}(*__args__, **__kwargs__)
    """

    context = sandbox.__dict__
    context["__args__"] = data["args"]
    context["__kwargs__"] = data["kwargs"]

    try:
        code_obj = compile(code, filename=data["filename"], mode="exec")
        exec(code_obj, context, results)
        ret = pickle.dumps(results["__return"])
    except Exception as exc:
        exc.__traceback_str__ = traceback.format_exc()  # type: ignore
        ret = pickle.dumps(exc)

    return ret


def inject(data: InjectData) -> bytes:
    result: bytes = call_function(data)

    return result


def get_source(project_dir: Path) -> Dict[str, str]:
    if not project_dir.exists():
        raise RuntimeError(f"{str(project_dir)} directory doesn't exist on the remote server.")

    source_code: Dict[str, str] = {}

    for p in glob.iglob(str(project_dir) + "/**/*.py", recursive=True):
        path: Path = Path(p)
        rel_path: str = str(path.relative_to(project_dir))
        source_code[rel_path] = path.read_text("utf-8")

    return source_code


def get_requirements(venv_path: Optional[Path] = None) -> Dict[str, Requirement]:
    paths: Optional[List[str]] = None
    if venv_path:
        site_packages = utils.get_site_packages_dir_from_venv(venv_path)
        paths = [str(site_packages)]

    cleared_reqs = {}

    from stickybeak.vendored.pip._vendor.pkg_resources import EggInfoDistribution

    for r in get_installed_distributions(
        paths=paths, skip=["pip", "pkg-resources", "setuptools", "packaging"], local_only=False, include_editables=False
    ):
        # Skip -e installs?
        if isinstance(r, EggInfoDistribution):
            continue

        req = Requirement(project_name=r.project_name, key=r.key, egg_info=r.egg_info, version=r.version)
        cleared_reqs[r.project_name] = req

    return cleared_reqs


def get_envs() -> Dict[str, str]:
    envs: Dict[str, str] = dict()

    for key, value in os.environ.items():
        envs[key] = value

    return envs


def get_server_data(project_dir: Union[str, Path, None]) -> ServerData:
    if project_dir:
        project_dir = Path(project_dir)
        source = get_source(project_dir)
    else:
        source = {}

    data = ServerData(
        source=source,
        requirements=get_requirements(),
        envs=get_envs(),
    )
    return data



from contextlib import closing
from dataclasses import dataclass, field
import hashlib
import os
from pathlib import Path
import re
import shutil
import socket
import subprocess
import sys
import textwrap
from threading import Thread
from time import sleep
import traceback
from typing import Any, Callable, Dict, List, Optional, Tuple
from urllib.parse import urljoin, urlparse

import dill as pickle
import requests.exceptions

from stickybeak import utils
from stickybeak.handle_requests import INJECT_ENDPOINT, SERVER_DATA_ENDPOINT, InjectData, Requirement, get_requirements
from stickybeak.utils import Client
from stickybeak.vendored import inspect, pip

__all__ = ["InjectorException", "Injector", "DjangoInjector", "FlaskInjector", "ConnectionError"]


class InjectorException(Exception):
    pass


class ConnectionError(Exception):
    pass


@dataclass
class DependencyInstallError(Exception):
    return_code: int


@dataclass
class BaseInjector:
    """Provides interface for code injection."""

    name: str  # injector name. Will be used as id in stickybeak_dir
    host: str
    port: int
    download_deps: bool
    stickybeak_dir: Path = field(init=False)  # directory where remote dependencies and project source is kept

    _client: Optional[Client] = field(init=False)

    _server_data: Dict[str, Any] = field(
        init=False, default_factory=dict
    )  # server data like source or pip freeze requirements

    connected: bool = field(init=False, default=False)

    def __post_init__(self) -> None:
        self._init()

    def _init(self) -> None:
        self._client = Client(self.host)
        self.stickybeak_dir = Path.home() / ".stickybeak" / Path(f"{self.name}")

    def connect(self, blocking: bool = True) -> None:
        def target() -> None:
            try:
                # ########## Get data
                self._server_data = self._client.get(SERVER_DATA_ENDPOINT)

                # ########## Collect remote code
                sources: Dict[str, str] = self._server_data["source"]

                for path, source in sources.items():
                    abs_path: Path = self.stickybeak_dir / Path(path)

                    abs_path.parent.mkdir(parents=True, exist_ok=True)
                    abs_path.touch()
                    abs_path.write_text(source, "utf-8")

                # ########## collect requirements
                if self.download_deps:
                    self._do_download_deps()

                self.connected = True
            except requests.exceptions.ConnectionError as e:
                raise ConnectionError from None

        if blocking:
            target()
        else:
            Thread(target=target).start()

    def wait_until_connected(self, timeout: float = 5.0) -> None:
        waited = 0.0
        one_sleep = 0.1
        while not self.connected:
            sleep(one_sleep)
            waited += one_sleep

            if waited >= timeout:
                raise TimeoutError

    def _do_download_deps(self) -> None:
        venv_dir = (self.stickybeak_dir / ".venv").absolute()

        if not venv_dir.exists():
            subprocess.check_output([f"virtualenv", f"{venv_dir}"], stderr=subprocess.DEVNULL)

        remote_reqs = self._server_data["requirements"]
        local_reqs = get_requirements(venv_dir)

        @dataclass
        class ReqDiff:
            local: Optional[Requirement]
            remote: Requirement

        reqs_diff = {}
        for p, v in remote_reqs.items():
            remote = remote_reqs[p]
            local = local_reqs.get(p)

            if not local or remote["version"] != local["version"]:
                reqs_diff[p] = ReqDiff(local=local, remote=remote)

        if reqs_diff:
            # delete packages manualy (sometimes pip doesn't downgrade for some reason)
            site_packages = utils.get_site_packages_dir_from_venv(venv_dir)
            for p, r in reqs_diff.items():
                if not r.local:
                    continue
                package_dir = site_packages / r.local["key"]
                shutil.rmtree(package_dir, ignore_errors=True)

                shutil.rmtree(r.local["egg_info"], ignore_errors=True)

            reqs = [f"{p}=={r.remote['version']}" for p, r in reqs_diff.items()]
            ret = pip.main(["install", f"--target={str(site_packages)}", "--upgrade", *reqs])

            if ret:
                raise DependencyInstallError(return_code=ret)

    def _raise_if_not_connected(self) -> None:
        if not self.connected:
            raise InjectorException("Injector not connected! Run connect() first.")

    def _run_remote_fun(self, source: str, filename: str, offset: int, call: str, args: Any, kwargs: Any) -> object:
        """Execute code.
        Returns:
            Dictionary containing all local variables.
        Raises:
            All exceptions from the code run remotely.
        """
        self._raise_if_not_connected()

        # we have to unload all the django modules so django accepts the new configuration
        # make a module copy so we can iterate over it and delete modules from the original one
        modules_before: List[str] = list(sys.modules.keys())[:]
        sys_path_before = sys.path[:]

        envs_before: os._Environ = os.environ.copy()  # type: ignore
        os.environ = self._server_data["envs"]  # type: ignore

        if self.download_deps:
            sys.path = [p for p in sys.path if "site-packages" not in p]

            # remove project dir from sys.path so there's no conflicts
            sys.path.pop(0)
            sys.path.insert(0, str(self.stickybeak_dir.absolute()))

            site_packages = utils.get_site_packages_dir_from_venv(self.stickybeak_dir.absolute() / ".venv")
            sys.path = [str(site_packages), *sys.path]

        self._before_execute()
        data = InjectData(source=source, filename=filename, offset=offset, call=call, args=list(args), kwargs=kwargs)
        pickled_data = pickle.dumps(data)
        try:
            content: bytes = self._client.post(INJECT_ENDPOINT, data=pickled_data).content
        except requests.exceptions.ConnectionError:
            raise ConnectionError from None
        ret: object = pickle.loads(content)

        os.environ = envs_before
        sys.path = sys_path_before

        modules_after: List[str] = list(sys.modules.keys())[:]
        diff: List[str] = list(set(modules_after) - set(modules_before))

        for m in diff:
            sys.modules.pop(m)

        if isinstance(ret, Exception):
            sys.stderr.write(ret.__traceback_str__)  # type: ignore
            raise ret

        return ret

    def _before_execute(self) -> None:
        pass

    def _get_fun_src(self, fun: Callable[[], None]) -> Tuple[str, int]:
        code: str = inspect.getsource(fun)
        offset = inspect.getsourcelines(fun)[1]

        code_lines: List[str] = code.splitlines(True)

        if "@" in code_lines[0]:
            code_lines.pop(0)
            offset += 1

        code = "".join(code_lines)

        # remove indent that's left
        code = textwrap.dedent(code)
        return code, offset

    def run_fun(self, fun: Callable, *args: Any, **kwargs: Any) -> object:
        self._raise_if_not_connected()

        source, offset = self._get_fun_src(fun)
        filename = inspect.getsourcefile(fun)
        ret = self._run_remote_fun(
            source, filename=filename, offset=offset, call=fun.__name__, args=args, kwargs=kwargs
        )
        return ret

    def run_klass_fun(
        self,
        klass: type,
        fun: Callable,
        args: Tuple[Any],
        kwargs: Dict[str, Any],
    ) -> object:
        self._raise_if_not_connected()

        source: str = textwrap.dedent(inspect.getsource(klass))
        filename = inspect.getsourcefile(klass)
        offset = inspect.getsourcelines(klass)[1]
        ret = self._run_remote_fun(
            source, filename=filename, offset=offset, call=f"{klass.__name__}.{fun.__name__}", args=args, kwargs=kwargs
        )
        return ret

    def klass(self, cls: type) -> type:
        # re execute class to get copy
        # this way original class can yield multiple injected classes
        # first_instance = injector1.klass(Klass)
        # second_instance = injector2.klass(Klass)
        source: str = textwrap.dedent(inspect.getsource(cls))

        definition_module = sys.modules[cls.__module__]
        sandbox: Dict[str, Any] = {}
        exec(source, definition_module.__dict__, sandbox)
        cls_cpy = sandbox[cls.__name__]

        cls_cpy._injector = self  # type: ignore
        methods: List[str] = [a for a in dir(cls_cpy) if not a.startswith("__") and callable(getattr(cls_cpy, a))]

        for m in methods:

            def decorator(func: Callable[[], None]) -> Callable:
                def wrapped(*args: Any, **kwargs: Any) -> object:
                    return cls_cpy._injector.run_klass_fun(cls_cpy, func, args, kwargs)  # type: ignore

                return wrapped

            method: Callable[[], None] = getattr(cls_cpy, m)
            setattr(cls_cpy, m, decorator(method))

        return cls_cpy

    def function(self, fun: Callable[[], None]) -> Callable:
        """
        Decorator
        :param fun: function to be decorated:
        :return decorated function:
        """

        def wrapped(*args: Any, **kwargs: Any) -> object:
            ret = self.run_fun(fun, *args, **kwargs)
            return ret

        return wrapped


@dataclass
class Injector(BaseInjector):
    port: int = field(init=False)

    def __post_init__(self) -> None:
        pass

    def _get_free_port(self) -> int:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def prepare(self, port: Optional[int] = None) -> None:
        self.port = port or self._get_free_port()
        self.host = f"{self.host}:{self.port}"
        self._init()


@dataclass
class DjangoInjector(BaseInjector):
    port: int = field(init=False)

    django_settings_module: str

    def _before_execute(self) -> None:
        if not self.download_deps:
            return
        modules = list(sys.modules.keys())[:]
        for m in modules:
            if "django" in m:
                sys.modules.pop(m)

        os.environ["DJANGO_SETTINGS_MODULE"] = self.django_settings_module
        import django

        django.setup()


@dataclass
class FlaskInjector(BaseInjector):
    port: int = field(init=False)



# Expose a limited set of classes and functions so callers outside of
# the vcs package don't need to import deeper than `pip._internal.vcs`.
# (The test directory and imports protected by MYPY_CHECK_RUNNING may
# still need to import from a vcs sub-package.)
# Import all vcs modules to register each VCS in the VcsSupport object.
import stickybeak.vendored.pip._internal.vcs.bazaar
import stickybeak.vendored.pip._internal.vcs.git
import stickybeak.vendored.pip._internal.vcs.mercurial
import stickybeak.vendored.pip._internal.vcs.subversion  # noqa: F401
from stickybeak.vendored.pip._internal.vcs.versioncontrol import (  # noqa: F401
    RemoteNotFoundError,
    is_url,
    make_vcs_requirement_url,
    vcs,
)



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from stickybeak.handle_requests import INJECT_ENDPOINT, SERVER_DATA_ENDPOINT

from . import views

urlpatterns = [
    path(INJECT_ENDPOINT, csrf_exempt(views.InjectView.as_view()), name="inject"),
    path(SERVER_DATA_ENDPOINT, csrf_exempt(views.DataView.as_view()), name="data"),
]



import json
from pathlib import Path

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.views.generic import View

import dill as pickle

from stickybeak.handle_requests import InjectData, get_server_data, inject


class InjectView(View):
    def post(self, request: HttpRequest) -> HttpResponse:
        data: InjectData = pickle.loads(request.body)
        return HttpResponse(inject(data), status=200)


class DataView(View):
    def get(self, request: HttpRequest) -> HttpResponse:
        project_dir: Path = Path(settings.BASE_DIR)
        data = get_server_data(project_dir)
        response_data = json.dumps(data)
        return HttpResponse(response_data, status=200)



"""Prepares a distribution for installation
"""

# The following comment should be removed at some point in the future.
# mypy: strict-optional=False

import logging
import mimetypes
import os
import shutil

from stickybeak.vendored.pip._vendor.packaging.utils import canonicalize_name
from stickybeak.vendored.pip._vendor.six import PY2

from stickybeak.vendored.pip._internal.distributions import (
    make_distribution_for_install_requirement,
)
from stickybeak.vendored.pip._internal.distributions.installed import InstalledDistribution
from stickybeak.vendored.pip._internal.exceptions import (
    DirectoryUrlHashUnsupported,
    HashMismatch,
    HashUnpinned,
    InstallationError,
    NetworkConnectionError,
    PreviousBuildDirError,
    VcsHashUnsupported,
)
from stickybeak.vendored.pip._internal.models.wheel import Wheel
from stickybeak.vendored.pip._internal.network.download import BatchDownloader, Downloader
from stickybeak.vendored.pip._internal.network.lazy_wheel import (
    HTTPRangeRequestUnsupported,
    dist_from_wheel_url,
)
from stickybeak.vendored.pip._internal.utils.filesystem import copy2_fixed
from stickybeak.vendored.pip._internal.utils.hashes import MissingHashes
from stickybeak.vendored.pip._internal.utils.logging import indent_log
from stickybeak.vendored.pip._internal.utils.misc import (
    display_path,
    hide_url,
    path_to_display,
    rmtree,
)
from stickybeak.vendored.pip._internal.utils.temp_dir import TempDirectory
from stickybeak.vendored.pip._internal.utils.typing import MYPY_CHECK_RUNNING
from stickybeak.vendored.pip._internal.utils.unpacking import unpack_file
from stickybeak.vendored.pip._internal.vcs import vcs

if MYPY_CHECK_RUNNING:
    from typing import Callable, Dict, Iterable, List, Optional, Tuple

    from mypy_extensions import TypedDict
    from stickybeak.vendored.pip._vendor.pkg_resources import Distribution

    from stickybeak.vendored.pip._internal.index.package_finder import PackageFinder
    from stickybeak.vendored.pip._internal.models.link import Link
    from stickybeak.vendored.pip._internal.network.session import PipSession
    from stickybeak.vendored.pip._internal.req.req_install import InstallRequirement
    from stickybeak.vendored.pip._internal.req.req_tracker import RequirementTracker
    from stickybeak.vendored.pip._internal.utils.hashes import Hashes

    if PY2:
        CopytreeKwargs = TypedDict(
            'CopytreeKwargs',
            {
                'ignore': Callable[[str, List[str]], List[str]],
                'symlinks': bool,
            },
            total=False,
        )
    else:
        CopytreeKwargs = TypedDict(
            'CopytreeKwargs',
            {
                'copy_function': Callable[[str, str], None],
                'ignore': Callable[[str, List[str]], List[str]],
                'ignore_dangling_symlinks': bool,
                'symlinks': bool,
            },
            total=False,
        )

logger = logging.getLogger(__name__);logger.disabled=True


def _get_prepared_distribution(
    req,  # type: InstallRequirement
    req_tracker,  # type: RequirementTracker
    finder,  # type: PackageFinder
    build_isolation,  # type: bool
):
    # type: (...) -> Distribution
    """Prepare a distribution for installation."""
    abstract_dist = make_distribution_for_install_requirement(req)
    with req_tracker.track(req):
        abstract_dist.prepare_distribution_metadata(finder, build_isolation)
    return abstract_dist.get_pkg_resources_distribution()


def unpack_vcs_link(link, location):
    # type: (Link, str) -> None
    vcs_backend = vcs.get_backend_for_scheme(link.scheme)
    assert vcs_backend is not None
    vcs_backend.unpack(location, url=hide_url(link.url))


class File(object):

    def __init__(self, path, content_type):
        # type: (str, Optional[str]) -> None
        self.path = path
        if content_type is None:
            self.content_type = mimetypes.guess_type(path)[0]
        else:
            self.content_type = content_type


def get_http_url(
    link,  # type: Link
    download,  # type: Downloader
    download_dir=None,  # type: Optional[str]
    hashes=None,  # type: Optional[Hashes]
):
    # type: (...) -> File
    temp_dir = TempDirectory(kind="unpack", globally_managed=True)
    # If a download dir is specified, is the file already downloaded there?
    already_downloaded_path = None
    if download_dir:
        already_downloaded_path = _check_download_dir(
            link, download_dir, hashes
        )

    if already_downloaded_path:
        from_path = already_downloaded_path
        content_type = None
    else:
        # let's download to a tmp dir
        from_path, content_type = download(link, temp_dir.path)
        if hashes:
            hashes.check_against_path(from_path)

    return File(from_path, content_type)


def _copy2_ignoring_special_files(src, dest):
    # type: (str, str) -> None
    """Copying special files is not supported, but as a convenience to users
    we skip errors copying them. This supports tools that may create e.g.
    socket files in the project source directory.
    """
    try:
        copy2_fixed(src, dest)
    except shutil.SpecialFileError as e:
        # SpecialFileError may be raised due to either the source or
        # destination. If the destination was the cause then we would actually
        # care, but since the destination directory is deleted prior to
        # copy we ignore all of them assuming it is caused by the source.
        logger.warning(
            "Ignoring special file error '%s' encountered copying %s to %s.",
            str(e),
            path_to_display(src),
            path_to_display(dest),
        )


def _copy_source_tree(source, target):
    # type: (str, str) -> None
    target_abspath = os.path.abspath(target)
    target_basename = os.path.basename(target_abspath)
    target_dirname = os.path.dirname(target_abspath)

    def ignore(d, names):
        # type: (str, List[str]) -> List[str]
        skipped = []  # type: List[str]
        if d == source:
            # Pulling in those directories can potentially be very slow,
            # exclude the following directories if they appear in the top
            # level dir (and only it).
            # See discussion at https://github.com/pypa/pip/pull/6770
            skipped += ['.tox', '.nox']
        if os.path.abspath(d) == target_dirname:
            # Prevent an infinite recursion if the target is in source.
            # This can happen when TMPDIR is set to ${PWD}/...
            # and we copy PWD to TMPDIR.
            skipped += [target_basename]
        return skipped

    kwargs = dict(ignore=ignore, symlinks=True)  # type: CopytreeKwargs

    if not PY2:
        # Python 2 does not support copy_function, so we only ignore
        # errors on special file copy in Python 3.
        kwargs['copy_function'] = _copy2_ignoring_special_files

    shutil.copytree(source, target, **kwargs)


def get_file_url(
    link,  # type: Link
    download_dir=None,  # type: Optional[str]
    hashes=None  # type: Optional[Hashes]
):
    # type: (...) -> File
    """Get file and optionally check its hash.
    """
    # If a download dir is specified, is the file already there and valid?
    already_downloaded_path = None
    if download_dir:
        already_downloaded_path = _check_download_dir(
            link, download_dir, hashes
        )

    if already_downloaded_path:
        from_path = already_downloaded_path
    else:
        from_path = link.file_path

    # If --require-hashes is off, `hashes` is either empty, the
    # link's embedded hash, or MissingHashes; it is required to
    # match. If --require-hashes is on, we are satisfied by any
    # hash in `hashes` matching: a URL-based or an option-based
    # one; no internet-sourced hash will be in `hashes`.
    if hashes:
        hashes.check_against_path(from_path)
    return File(from_path, None)


def unpack_url(
    link,  # type: Link
    location,  # type: str
    download,  # type: Downloader
    download_dir=None,  # type: Optional[str]
    hashes=None,  # type: Optional[Hashes]
):
    # type: (...) -> Optional[File]
    """Unpack link into location, downloading if required.

    :param hashes: A Hashes object, one of whose embedded hashes must match,
        or HashMismatch will be raised. If the Hashes is empty, no matches are
        required, and unhashable types of requirements (like VCS ones, which
        would ordinarily raise HashUnsupported) are allowed.
    """
    # non-editable vcs urls
    if link.is_vcs:
        unpack_vcs_link(link, location)
        return None

    # If it's a url to a local directory
    if link.is_existing_dir():
        if os.path.isdir(location):
            rmtree(location)
        _copy_source_tree(link.file_path, location)
        return None

    # file urls
    if link.is_file:
        file = get_file_url(link, download_dir, hashes=hashes)

    # http urls
    else:
        file = get_http_url(
            link,
            download,
            download_dir,
            hashes=hashes,
        )

    # unpack the archive to the build dir location. even when only downloading
    # archives, they have to be unpacked to parse dependencies, except wheels
    if not link.is_wheel:
        unpack_file(file.path, location, file.content_type)

    return file


def _check_download_dir(link, download_dir, hashes):
    # type: (Link, str, Optional[Hashes]) -> Optional[str]
    """ Check download_dir for previously downloaded file with correct hash
        If a correct file is found return its path else None
    """
    download_path = os.path.join(download_dir, link.filename)

    if not os.path.exists(download_path):
        return None

    # If already downloaded, does its hash match?
    logger.info('File was already downloaded %s', download_path)
    if hashes:
        try:
            hashes.check_against_path(download_path)
        except HashMismatch:
            logger.warning(
                'Previously-downloaded file %s has bad hash. '
                'Re-downloading.',
                download_path
            )
            os.unlink(download_path)
            return None
    return download_path


class RequirementPreparer(object):
    """Prepares a Requirement
    """

    def __init__(
        self,
        build_dir,  # type: str
        download_dir,  # type: Optional[str]
        src_dir,  # type: str
        wheel_download_dir,  # type: Optional[str]
        build_isolation,  # type: bool
        req_tracker,  # type: RequirementTracker
        session,  # type: PipSession
        progress_bar,  # type: str
        finder,  # type: PackageFinder
        require_hashes,  # type: bool
        use_user_site,  # type: bool
        lazy_wheel,  # type: bool
    ):
        # type: (...) -> None
        super(RequirementPreparer, self).__init__()

        self.src_dir = src_dir
        self.build_dir = build_dir
        self.req_tracker = req_tracker
        self._session = session
        self._download = Downloader(session, progress_bar)
        self._batch_download = BatchDownloader(session, progress_bar)
        self.finder = finder

        # Where still-packed archives should be written to. If None, they are
        # not saved, and are deleted immediately after unpacking.
        self.download_dir = download_dir

        # Where still-packed .whl files should be written to. If None, they are
        # written to the download_dir parameter. Separate to download_dir to
        # permit only keeping wheel archives for pip wheel.
        self.wheel_download_dir = wheel_download_dir

        # NOTE
        # download_dir and wheel_download_dir overlap semantically and may
        # be combined if we're willing to have non-wheel archives present in
        # the wheelhouse output by 'pip wheel'.

        # Is build isolation allowed?
        self.build_isolation = build_isolation

        # Should hash-checking be required?
        self.require_hashes = require_hashes

        # Should install in user site-packages?
        self.use_user_site = use_user_site

        # Should wheels be downloaded lazily?
        self.use_lazy_wheel = lazy_wheel

        # Memoized downloaded files, as mapping of url: (path, mime type)
        self._downloaded = {}  # type: Dict[str, Tuple[str, str]]

    @property
    def _download_should_save(self):
        # type: () -> bool
        if not self.download_dir:
            return False

        if os.path.exists(self.download_dir):
            return True

        logger.critical('Could not find download directory')
        raise InstallationError(
            "Could not find or access download directory '{}'"
            .format(self.download_dir))

    def _log_preparing_link(self, req):
        # type: (InstallRequirement) -> None
        """Log the way the link prepared."""
        if req.link.is_file:
            path = req.link.file_path
            logger.info('Processing %s', display_path(path))
        else:
            logger.info('Collecting %s', req.req or req)

    def _get_download_dir(self, link):
        # type: (Link) -> Optional[str]
        if link.is_wheel and self.wheel_download_dir:
            # Download wheels to a dedicated dir when doing `pip wheel`.
            return self.wheel_download_dir
        return self.download_dir

    def _ensure_link_req_src_dir(self, req, download_dir, parallel_builds):
        # type: (InstallRequirement, Optional[str], bool) -> None
        """Ensure source_dir of a linked InstallRequirement."""
        # Since source_dir is only set for editable requirements.
        if req.link.is_wheel:
            # We don't need to unpack wheels, so no need for a source
            # directory.
            return
        assert req.source_dir is None
        # We always delete unpacked sdists after pip runs.
        req.ensure_has_source_dir(
            self.build_dir,
            autodelete=True,
            parallel_builds=parallel_builds,
        )

        # If a checkout exists, it's unwise to keep going.  version
        # inconsistencies are logged later, but do not fail the
        # installation.
        # FIXME: this won't upgrade when there's an existing
        # package unpacked in `req.source_dir`
        if os.path.exists(os.path.join(req.source_dir, 'setup.py')):
            raise PreviousBuildDirError(
                "pip can't proceed with requirements '{}' due to a"
                "pre-existing build directory ({}). This is likely "
                "due to a previous installation that failed . pip is "
                "being responsible and not assuming it can delete this. "
                "Please delete it and try again.".format(req, req.source_dir)
            )

    def _get_linked_req_hashes(self, req):
        # type: (InstallRequirement) -> Hashes
        # By the time this is called, the requirement's link should have
        # been checked so we can tell what kind of requirements req is
        # and raise some more informative errors than otherwise.
        # (For example, we can raise VcsHashUnsupported for a VCS URL
        # rather than HashMissing.)
        if not self.require_hashes:
            return req.hashes(trust_internet=True)

        # We could check these first 2 conditions inside unpack_url
        # and save repetition of conditions, but then we would
        # report less-useful error messages for unhashable
        # requirements, complaining that there's no hash provided.
        if req.link.is_vcs:
            raise VcsHashUnsupported()
        if req.link.is_existing_dir():
            raise DirectoryUrlHashUnsupported()

        # Unpinned packages are asking for trouble when a new version
        # is uploaded.  This isn't a security check, but it saves users
        # a surprising hash mismatch in the future.
        # file:/// URLs aren't pinnable, so don't complain about them
        # not being pinned.
        if req.original_link is None and not req.is_pinned:
            raise HashUnpinned()

        # If known-good hashes are missing for this requirement,
        # shim it with a facade object that will provoke hash
        # computation and then raise a HashMissing exception
        # showing the user what the hash should be.
        return req.hashes(trust_internet=False) or MissingHashes()

    def _fetch_metadata_using_lazy_wheel(self, link):
        # type: (Link) -> Optional[Distribution]
        """Fetch metadata using lazy wheel, if possible."""
        if not self.use_lazy_wheel:
            return None
        if self.require_hashes:
            logger.debug('Lazy wheel is not used as hash checking is required')
            return None
        if link.is_file or not link.is_wheel:
            logger.debug(
                'Lazy wheel is not used as '
                '%r does not points to a remote wheel',
                link,
            )
            return None

        wheel = Wheel(link.filename)
        name = canonicalize_name(wheel.name)
        logger.info(
            'Obtaining dependency information from %s %s',
            name, wheel.version,
        )
        url = link.url.split('#', 1)[0]
        try:
            return dist_from_wheel_url(name, url, self._session)
        except HTTPRangeRequestUnsupported:
            logger.debug('%s does not support range requests', url)
            return None

    def prepare_linked_requirement(self, req, parallel_builds=False):
        # type: (InstallRequirement, bool) -> Distribution
        """Prepare a requirement to be obtained from req.link."""
        assert req.link
        link = req.link
        self._log_preparing_link(req)
        with indent_log():
            wheel_dist = self._fetch_metadata_using_lazy_wheel(link)
        if wheel_dist is not None:
            req.needs_more_preparation = True
            return wheel_dist
        return self._prepare_linked_requirement(req, parallel_builds)

    def prepare_linked_requirements_more(self, reqs, parallel_builds=False):
        # type: (Iterable[InstallRequirement], bool) -> None
        """Prepare a linked requirement more, if needed."""
        reqs = [req for req in reqs if req.needs_more_preparation]
        links = []  # type: List[Link]
        for req in reqs:
            download_dir = self._get_download_dir(req.link)
            if download_dir is not None:
                hashes = self._get_linked_req_hashes(req)
                file_path = _check_download_dir(req.link, download_dir, hashes)
            if download_dir is None or file_path is None:
                links.append(req.link)
            else:
                self._downloaded[req.link.url] = file_path, None

        # Let's download to a temporary directory.
        tmpdir = TempDirectory(kind="unpack", globally_managed=True).path
        self._downloaded.update(self._batch_download(links, tmpdir))
        for req in reqs:
            self._prepare_linked_requirement(req, parallel_builds)

    def _prepare_linked_requirement(self, req, parallel_builds):
        # type: (InstallRequirement, bool) -> Distribution
        assert req.link
        link = req.link
        download_dir = self._get_download_dir(link)

        with indent_log():
            self._ensure_link_req_src_dir(req, download_dir, parallel_builds)
            hashes = self._get_linked_req_hashes(req)
            if link.url not in self._downloaded:
                try:
                    local_file = unpack_url(
                        link, req.source_dir, self._download,
                        download_dir, hashes,
                    )
                except NetworkConnectionError as exc:
                    raise InstallationError(
                        'Could not install requirement {} because of HTTP '
                        'error {} for URL {}'.format(req, exc, link)
                    )
            else:
                file_path, content_type = self._downloaded[link.url]
                if hashes:
                    hashes.check_against_path(file_path)
                local_file = File(file_path, content_type)

            # For use in later processing, preserve the file path on the
            # requirement.
            if local_file:
                req.local_file_path = local_file.path

            dist = _get_prepared_distribution(
                req, self.req_tracker, self.finder, self.build_isolation,
            )

            if download_dir:
                if link.is_existing_dir():
                    logger.info('Link is a directory, ignoring download_dir')
                elif local_file:
                    download_location = os.path.join(
                        download_dir, link.filename
                    )
                    if not os.path.exists(download_location):
                        shutil.copy(local_file.path, download_location)
                        download_path = display_path(download_location)
                        logger.info('Saved %s', download_path)

            if self._download_should_save:
                # Make a .zip of the source_dir we already created.
                if link.is_vcs:
                    req.archive(self.download_dir)
        return dist

    def prepare_editable_requirement(
        self,
        req,  # type: InstallRequirement
    ):
        # type: (...) -> Distribution
        """Prepare an editable requirement
        """
        assert req.editable, "cannot prepare a non-editable req as editable"

        logger.info('Obtaining %s', req)

        with indent_log():
            if self.require_hashes:
                raise InstallationError(
                    'The editable requirement {} cannot be installed when '
                    'requiring hashes, because there is no single file to '
                    'hash.'.format(req)
                )
            req.ensure_has_source_dir(self.src_dir)
            req.update_editable(not self._download_should_save)

            dist = _get_prepared_distribution(
                req, self.req_tracker, self.finder, self.build_isolation,
            )

            if self._download_should_save:
                req.archive(self.download_dir)
            req.check_if_exists(self.use_user_site)

        return dist

    def prepare_installed_requirement(
        self,
        req,  # type: InstallRequirement
        skip_reason  # type: str
    ):
        # type: (...) -> Distribution
        """Prepare an already-installed requirement
        """
        assert req.satisfied_by, "req should have been satisfied but isn't"
        assert skip_reason is not None, (
            "did not get skip reason skipped but req.satisfied_by "
            "is set to {}".format(req.satisfied_by)
        )
        logger.info(
            'Requirement %s: %s (%s)',
            skip_reason, req, req.satisfied_by.version
        )
        with indent_log():
            if self.require_hashes:
                logger.debug(
                    'Since it is already installed, we are trusting this '
                    'package without checking its hash. To ensure a '
                    'completely repeatable environment, install into an '
                    'empty virtualenv.'
                )
            return InstalledDistribution(req).get_pkg_resources_distribution()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stickybeak',
 'stickybeak.dj',
 'stickybeak.vendored',
 'stickybeak.vendored.pip',
 'stickybeak.vendored.pip._internal',
 'stickybeak.vendored.pip._internal.cli',
 'stickybeak.vendored.pip._internal.commands',
 'stickybeak.vendored.pip._internal.distributions',
 'stickybeak.vendored.pip._internal.index',
 'stickybeak.vendored.pip._internal.models',
 'stickybeak.vendored.pip._internal.network',
 'stickybeak.vendored.pip._internal.operations',
 'stickybeak.vendored.pip._internal.operations.build',
 'stickybeak.vendored.pip._internal.operations.install',
 'stickybeak.vendored.pip._internal.req',
 'stickybeak.vendored.pip._internal.resolution',
 'stickybeak.vendored.pip._internal.resolution.legacy',
 'stickybeak.vendored.pip._internal.resolution.resolvelib',
 'stickybeak.vendored.pip._internal.utils',
 'stickybeak.vendored.pip._internal.vcs',
 'stickybeak.vendored.pip._vendor',
 'stickybeak.vendored.pip._vendor.cachecontrol',
 'stickybeak.vendored.pip._vendor.cachecontrol.caches',
 'stickybeak.vendored.pip._vendor.certifi',
 'stickybeak.vendored.pip._vendor.chardet',
 'stickybeak.vendored.pip._vendor.chardet.cli',
 'stickybeak.vendored.pip._vendor.colorama',
 'stickybeak.vendored.pip._vendor.distlib',
 'stickybeak.vendored.pip._vendor.distlib._backport',
 'stickybeak.vendored.pip._vendor.html5lib',
 'stickybeak.vendored.pip._vendor.html5lib._trie',
 'stickybeak.vendored.pip._vendor.html5lib.filters',
 'stickybeak.vendored.pip._vendor.html5lib.treeadapters',
 'stickybeak.vendored.pip._vendor.html5lib.treebuilders',
 'stickybeak.vendored.pip._vendor.html5lib.treewalkers',
 'stickybeak.vendored.pip._vendor.idna',
 'stickybeak.vendored.pip._vendor.msgpack',
 'stickybeak.vendored.pip._vendor.packaging',
 'stickybeak.vendored.pip._vendor.pep517',
 'stickybeak.vendored.pip._vendor.pkg_resources',
 'stickybeak.vendored.pip._vendor.progress',
 'stickybeak.vendored.pip._vendor.requests',
 'stickybeak.vendored.pip._vendor.resolvelib',
 'stickybeak.vendored.pip._vendor.resolvelib.compat',
 'stickybeak.vendored.pip._vendor.toml',
 'stickybeak.vendored.pip._vendor.urllib3',
 'stickybeak.vendored.pip._vendor.urllib3.contrib',
 'stickybeak.vendored.pip._vendor.urllib3.contrib._securetransport',
 'stickybeak.vendored.pip._vendor.urllib3.packages',
 'stickybeak.vendored.pip._vendor.urllib3.packages.backports',
 'stickybeak.vendored.pip._vendor.urllib3.packages.ssl_match_hostname',
 'stickybeak.vendored.pip._vendor.urllib3.util',
 'stickybeak.vendored.pip._vendor.webencodings']

package_data = \
{'': ['*']}

install_requires = \
['dill', 'requests', 'typing-extensions', 'virtualenv']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.9']}

setup_kwargs = {
    'name': 'stickybeak',
    'version': '0.7.9',
    'description': 'Package that makes e2e tests easy.',
    'long_description': None,
    'author': 'Damian Krystkiewicz',
    'author_email': 'damian.krystkiewicz@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/reloadware/stickybeak',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
